      program frt23
c   Forest reflectance and transmittance, and its inversion.
c   (C) Andres Kuusk & Tiit Nilson
c   Kuusk, A. and Nilson, T. A directional multispectral forest 
c   reflectance model, Remote Sens. Environ., 2000, 72(2): 
c   244-252.
c   Version 12.2024.
c   Contact: A. Kuusk, Tartu Observatory, Estonia, andres@bgf.ee
c
c   December_2023-November_2024: Major revision of the code structure,
c   The format of input files is changed.
c   The loops (ijob=1 and ijob=2) are organized in the main program.
c
c
c   April 2013: Added the inversion option ijob=5 - a single wavelenght
c   and single sun zenith angle but several view angles.
c
c   Januar 2023 added the warning: 
c   *** Warning!: canopy closure (CC) > crown closure (CR)
c     If canopy closure is greater than crown closure 
c     then the regularity parameter "tree  distr.  param    13" 
c     for this class is too large. 
c
c   January 2013: The phase function Henyey-Greenstein is used for 
c   pine and spruce. The asymmetry parameter g (ghg in the code)
c   is fixed in the beginning of the main program:
c     ghg = 0.5d0  ! RAMI spruce, f = 0.829
c     ghg = 0.4d0  ! RAMI pine, f = 0.775, 
c     f - the asymmetry of the phase function
c
c   November-December 2012: Added the LAD of leaves using the 
c   approximations for the G-function and gamma of elliptical LAD.
c   The homogeneous model ACRMf for the ground vegetation.
c   The structure of input file changes.
c
c   July 2009: Abdelaziz Kallel detected an error in considering
c   soil component in diffuse fluxes, updated dif3.f.
c   In layer.f added the approximation for small absorption.
c
c   May 2006: Modified the calculation of diffuse fluxes in modules
c   dif3() and diffor(). Changed the read-directives in frt.f
c   in order to be usable both in Unix and DOS.
c
c   Aprill 2005: Added the shape of the tree stem.
c   The regularity of stem positions is not accounted for.
c   The code is compiled using Fortran95.
c
c   January 2003: Replaced cubatures for the ellipsoid
c   cubell(5) -> cubell9
c
c   September 2002: Calculation of diffuse fluxes with Verhoef model.
c   Some errors of downward radiances of crown layer fixed.
c
c   August 2002: The 3D integration using cubatures.
c
c   February 2002: Added the option of leaf optics model:
c   PROSPECT or Liberty. NB! Liberty may have problems.
c   The 2-layer homogeneous model for the ground vegetation.
c   The number of leaf biochemistry components increased up to 10.
c
c   December 2001: Modified error messages, 
c   some small corrections in the code.
c
c   May 2001: Replaced the PROSPECT model to the 1997 version.
c
c   April 2001: Modified the merit function, added the expert guess.
c
c   December 2000: Merged the separate programs ifrt, frtspec, frtad,
c   the structure of the input file changed.
c   Input and output are in the main code frt.f, calculations
c   are organised in the module funcf.f
c
c   September 1999: The hot-spot parameter separately
c   for every tree class. Modified the structure of the input file.
c
c   October 1998: Diffuse reflectance is calculated 
c   integrating over polar angles at phi = 90 deg.
c
c   July 1998: Added single scattering on tree stems.
c
c   May 1998: Model works for every view direction. The same code
c   for the direct run and inversion.
c
c   Märts 1998: Metsa läbipaistvus binoomvalemiga,
c   mitmekordse hajumise valemis efektiivne LAI Cheni 
c   valemitega, alustaimestiku heleduskoefitsient 
c   homogeense taimkatte mudeliga MCRM.
c
c   Märts, 1996: alustaimestiku peegeldumisspekter failis,
c   Price'i vektoreid otseselt ei kasuta, seega ilmutatud
c   parameetrite arv kahanes 14-le.
c
c   A. Kuusk 23.10.1995 Avignonis
c
c   August, 1995 Grignonis: algandmetes ka Q/D suhe, 
c   Pöördülesandes saab kasutada nii heleduskoefitsiendi
c   kui läbilaskvuskoefitsientide väärtusi. Vajaduse korral
c   tuleb muuta sihtfunktsiooni arvutamist funcifrt.F-is
c
      include 'par.h'
      include 'stands.h'
      include 'spectra.h'
      include 'func.h'
      include 'volint.h'
c
      logical lig, lml
      character fname*72, xfs*16
      character*2 trl
      character*6 mv
      character*40 dname
      character outx*72
c
      dimension dcaer(4)
      dimension llxx(nvar), xmin(nvar), xmax(nvar), xx(nvar), 
     &      x0(nvar), xxf(nvar), xdelt(nvar),  xa(nvar), xb(nvar)
      dimension xc(nclmax, nvar)
      dimension parr(nvar+1, nvar)
c
      common /pidr/ pi, dr
      common /par/  zeps, tol, tiny, ftol, alphopt, betopt, gammopt, 
     & at, dxopt, ier, ncount, nfmax, itmax, itbr, nbrak, ilo
c
      data lig/.false./
      data trl/'# '/
      data elli/'  ellips'/, koon/'  cone+c'/
      data eps/.1d-2/, eps2/1.745d-2/, thtv0/0.d0/
c
*     call fptrap(1)
c
      pi = acos(-1.d0)
      dr = pi/180.d0
c
      iwr    = 1
      ncount = 0
*     nrefl  = 0
      ier    = 0
      bangs  = -1.d0
      pkhair = .1d0
*     dcell0 = 80.d0
      nzs    = 4
      nxx  = 1
c
      call getarg(1, fname)
      call getarg(2, oufile)
      if (fname .eq. ' ') then
         write (0, '(a)')  ' Name of input file '
         read (*, '(a)')     fname
      endif
      if (oufile .eq. ' ') then
         write (0, '(a)')  ' Name of output file '
         read (*, '(a)')     oufile
      endif
      clfile(1) = fname
c                                                   Read input files
      call stands
     & (ijob, ncl, nbnds, nxx, llxx, iaer, dcaer, dvis, 
     & dtaer55, thsd, thvd, dthvd, phid, xc, xmin, xmax, xdelt, 
     & dwl, frefr, fprice, lig, mv, dname, oufile) 
c
c                                                  Print input data
      call out0
     & (ijob, ncl, nbnds, nxx, llxx, iaer, dcaer, dvis, 
     & dtaer55, thsd, thvd, dthvd, phid, xc, xmin, xmax, xdelt, 
     & dwl, frefr, fprice, lig, mv, dname, oufile) 
c
c   Initialisation: the knots and weigths of Gauss-Legendre quadrature
      x1      = 0.d0
      x2      = pi/2.d0
      call gauleg(x1, x2, zctst, acztst, nknots)
      do jthk = 1, nknots
         thknot(jthk) = zctst(jthk)
         p14(jthk)    = cos(thknot(jthk))*acztst(jthk)
      enddo
      ntstz   = 2*nzs
      x1 = -1.d0
      x2 = 1.d0
      call gauleg(x1, x2, xptst, aptst, nknots)
      call gauleg(x1, x2, zctst, acztst, ntstz)
      call cubell9(netst, xetst, yetst, zetst, aetst)
      call cubcirc(nctst, xctst, yctst, actst)
c
      do jcl = 1, ncl
        if (lelli(jcl)) then
          kuju(jcl) = elli
        else
          kuju(jcl) = koon
        endif
      enddo
c                  ! xxf(..) is the argument of func(..) in the inversion
      do j = 1, nvar
        xxf(j) = 0.d0
      enddo
c
c  ******************                    ! Start of computations
c
c                                        ! Check wavelength range
      do jbnd = 1, nbnds
        if (dwl .gt. 0.d0) wlarr(jbnd) = wlarr(1) + (jbnd - 1)*dwl
        rlambda   = wlarr(jbnd)
        if ((rlambda .lt. 400.d0) .or. (rlambda .gt. 2400.d0)) 
     &    goto 88
      enddo
c
c                                            ! Check crown length
      do jcl = 1, ncl
        if (lelli(jcl)) then
          if (xc(jcl, 3) .gt. xc(jcl, 2)) then
            xc(jcl, 3) = xc(jcl, 2)
            print *, 'jcl = ', jcl
            print *, 'Warning: tree height is less than crown length,'
            print *, 'the crown length is adjusted.'
          endif
        else
          hcrown = xc(jcl, 3) + xc(jcl, 4)
          if (xc(jcl, 2) .lt. xc(jcl, 3)) then
            xc(jcl, 3) = xc(jcl, 2)
            xc(jcl, 4) = 0.d0
            print *, 'jcl = ', jcl
            print *, 'Warning: tree height is less than crown length,'
            print *, 'the crown length is adjusted.'
          else if (hcrown .gt. xc(jcl, 2)) then
            xc(jcl, 4) = xc(jcl, 2) - xc(jcl, 3)
            print *, 'jcl = ', jcl
            print *, 'Warning: tree height is less than crown length,'
            print *, 'the crown length is adjusted.'
          endif
        endif
      enddo
c
      thv = thvd*dr          ! view and sun angles to radians
      phi = phid*dr
      ths = thsd*dr
c                                              ! S/D ratio (6S)
      if (ijob .eq. 0 .or. ijob .eq. 2) then
        nbnds = 1
      endif
      jpass = 1
      do jbnd = 1, nbnds
        rlambda = wlarr(jbnd)
        jwl = nint((rlambda - (400.d0 - wlstep))/wlstep)
        if (iaer .eq. -1) then
          sqa(jwl) = 1.d0
        else
          call sixd
     &      (jpass, iwr, oufile, rlambda, 
     &         thss, iaer, dcaer, dvis, dtaer55, sqjwl)
          sqa(jwl) = sqjwl
        endif
        jpass = jpass + 1
      enddo
c
c                            ! copy parameters into common block
      l_opt   = .true.
      l_struc = .true.
      i_job   = ijob
c
      n_cl    = ncl
      n_zs    = nzs
c
c
      do jcl = 1, ncl
        do jx = 1, nvar
          x_c(jcl, jx)     = xc(jcl, jx)
        enddo
      enddo
c              initialize the vector of parameters - the main tree class 
c              in the direct run, and the initial guess in the inversion
      do jx = 1, nvar
        xcj = xc(1, jx)
        xx(jx) = xcj
        x0(jx) = xcj
      enddo
c                            ! copy parameters into common block(s)
      do jx = 1, nvar
        x_0(jx)    = x0(jx)
        x_x(jx)    = xx(jx)
      enddo
c
c                ******************  ijob = 0, the single direct run
      if (ijob .eq. 0) then
c
        jpass = 1
        jview = 1
        jbnd = 1
        rlambda = wlarr(jbnd)
        fff  = func(nxx, xxf)
        goto 770
      endif
c
c                ******************  Loop to calculate spectrum, ijob = 1
c
      if (ijob .eq. 1) then
        write (0, '(3(a, f4.1), i8, a/a)')
     &      ' th_Sun = ', thsd, '    thv = ', thvd,
     &      '    phi = ', phid, nbnds, ' spectral bands',
     &      ' wl          rlambda'
        jview = 1
        jpass = 0
        thv    = thvd*dr
        ths    = thsd*dr
        phi    = phid*dr
c
        fff  = func(nxx, xxf)
c
        goto 770
      endif
c                              ********** End of the spectrum loop
c
c            **********  Loop to calculate angular distribution, ijob = 2
c
      if (ijob .eq. 2) then
        jbnd = 1
        rlambda = wlarr(jbnd)
        thtv00 = 80.d0
        nview1  = int(thtv00/dthvd)
        nview   = 2*nview1 + 1
        if (abs(phi - pi/2.d0) .lt. eps) then
           nv = nview1 + 1
        else
           nv = nview
        endif
        dthv = dthvd*dr
        thtv0  = -(thtv00/dthvd + 1)*dthv
        write (*, '(2(a, f4.1), a, f6.1, i8, a)')
     &      '# th_Sun = ', thsd, '    phi = ', phid, 
     &      '   Spectral band ', rlambda, nview, ' view angles'
c
        do jv = 1, nv
          thvj   = thtv0 + jv*dthv
          if (abs(thvj) .lt. eps2) then
            thv   = 0.d0
            phi  = 0.d0
          endif
          if (jv .le. nview1) then
            thv = abs(thvj)
            phi = phid*dr
          else
            thv = thvj
            phi = pi - phid*dr
          endif
c                          ! view angle 
          jview  = jv
          l_struc = .true.
c
          fff  = func(nxx, xxf)
c
          if (abs(phi - pi/2.d0) .lt. eps) then
            jv2 = nview - jv
            rth(jv2)    = rth(jv)
            rsth(jv2)   = rsth(jv)
            bdown(jv2)    = bdown(jv)
            aavth(jv2)    = aavth(jv)
            p11gr(jv2)    = p11gr(jv)
            p11lvs(jv2)   = p11lvs(jv)
            rcr1(jv2)     = rcr1(jv)
            rgr1(jv2)     = rgr1(jv)
            rdif(jv2)     = rdif(jv)
          endif
*     write (*, '(f6.1, 2f8.4)') thv/dr, rcr1(jv), p11lvs(jv)
        enddo
        goto 770
c                            endif (ijob .eq. 2)
      endif
c
c                                ********* Inversion ********
c
      i_job = ijob
      jv = 1
      jl = 1
      ipass = 0
      ier = 0
      xfs    = oufile
      outx   = xfs//'0.dat'
      thv    = thvd*dr
      ths    = thsd*dr
      phi    = phid*dr
c
      fname = 'flow.dat'
      open (4, file = fname, status = 'old', err = 99)
      rewind 4
c
*     if (ijob .eq. 5) then
*        jths = 1
*        do jr = 1, nrefl
*           t_hvr(jr)          = thvr(jr)*dr
*           p_hir(jr)          = (180.d0 - abs(phir(jr)))*dr
*           b_qjarr(jths, jr)  = bqjarr(jths, jr)
*           d_rhoarr(jths, jr) = bqjarr(jths, jr)*0.1d0
*        enddo
*     endif
c
500   continue
      if (lig) then
c                          ! read the results of previous run
        fname = outx
        open (7, file = fname, status = 'old', err = 99)
        rewind 7
        read (7, '(d27.17)', err = 98) (xx(jx), jx = 1, nvar)
        close (7)
        do jv = 1, nvar
          xc(1, jv) = xx(jv)
        enddo 
      endif
c                            scale the vector of parameters
      do i = 1, nxx
        li        = llxx(i)
        xa(li)    = (xmax(li) + xmin(li))*.5d0
        xb(li)    = 10.d0/(xmax(li) - xmin(li))
        xxf(i)    = (xx(li) - xa(li))*xb(li)
        if (li .le. 13) then
          l_struc = .true.
        endif
      enddo
c                            ! copy parameters into common block(s)
      do jx = 1, nvar
        l_lxx(jx)  = llxx(jx)
        x_0(jx)    = xx(jx)
        x_x(jx)    = xx(jx)
        x_a(jx)    = xa(jx)
        x_b(jx)    = xb(jx)
        x_delt(jx) = xdelt(jx)
      enddo
c
      if (ipass .eq. 0) then
        fff  = func(nxx, xxf)
        write (*, '(/a, e20.14, a, i5, a, i3)')
     &    '  f(x0) = ', fff, '   ncount = ', ncount, '  ier = ', ier
        open (iwr, file = oufile, status = 'old')
        do j = 1, 6000
           read (iwr, '(1x)', end = 65)
        enddo
65      continue
        backspace iwr
c
        write (iwr, '(a/a/a/a, i3/a)') trl, trl,
     &    '########################', '#  Results,     pass', ipass,
     &    '########################'
*       write (iwr, '(a/a, g20.14)') trl, '#  f(x0) = ', fff
        write (iwr, '(/a/a, e20.14, a, i5, a, i3, a, f7.5)')
     &    trl, '# f = ', fff, '   ncount = ', ncount, '  ier = ', ier
c
        close (iwr)
      endif
      ipass = ipass + 1
      ier = 0
c
c                                    ***     powell
      if (mv .eq. 'powell') then
        do i = 1, nxx
          do j = 1, nxx
            if (i .eq. j) then
              parr(j, i) = dxopt
            else
              parr(j, i) = 0.d0
            endif
          enddo
        enddo
c
c
        call powell(xxf, parr, nxx, fff, iter)
c
        do il = 1, nxx
          li     = llxx(il)
          xx(li) = xxf(il)/xb(li) + xa(li)
          xc(1, li) = xx(li)
        enddo
      endif
      goto 772
c
**************************
c                               ! Output results, direct run
770   continue
      call out1
     & (ijob, ncl, nbnds, nview, 
     & thsd, thvd, dthvd, phid, thtv00, xc, dname, oufile)
      if (ijob .lt. 3) goto 776
c                                         ! Output results, inversion
772   continue
c                                         ! save the vector of parameters
      open (7, file = outx, status = 'unknown')
         rewind 7
         write (7, '(d27.17)') (xx(i), i = 1, nvar) 
      close (7)
c
*     print *, char(7)
c
      call out2
     & (ijob, ncl, nbnds, nxx, llxx, ipass, iter, xc, xmin, xmax, 
     & xdelt, xx, fff, dname, oufile)
c
      if (ier .eq. 1) then
        goto 776
      else
        read (4, *, err = 98) jj
         jl = jl + 1
         if (jj .eq. 7 .or. ncount .ge. nfmax) then
           goto 776
         else if (jj .eq. 1) then
           goto 500
         else
           print *, 'Error in the file ''flow.dat'''
           goto 776
         endif
      endif
c
776   continue
      close (4)
      close (iwr)
*     print *, char(7)
c
      stop 'FRT23'
c
c     ***********************
c     ***********************
c                                ! Error messages
88    continue
      jwl = nint((rlambda - (400.d0 - wlstep))/wlstep)
      print *, 'MSFR: Wavelength out of range'
      write (*, '(a, i6, g15.4)')
     &         'jw, lambda:', jwl, rlambda
      write (iwr, '(a, i6, g15.4 /)')
     &         'jw, lambda:', jwl, rlambda
      write (iwr, '(/)')
      close (iwr)
      stop 'FRT23'
c
98    write (*, '(3a, i4)') 
     &      'Error in the file ', fname, 'line ', jl
      stop 'FRT23'
99    write (*, '(a)')  'File not found: ', fname
      stop 'FRT23'
      end
*
************************************************************************
*
